<?php

/**
 * CDNBULL_Backups
 *
 * Singleton class for creating backups, all scheduling is handled by cdnbull
 */
class CDNBULL_Backups extends CDNBULL_HM_Backup {

	/**
	 * Contains the current instance
	 *
	 * @static
	 * @access private
	 */
	private static $instance;

	protected $filesize_transient;

	/**
	 * Return the current instance of CDNBULL_Backups
	 *
	 * @static
	 * @access public
	 */
	public static function get_instance() {

		if ( empty( self::$instance ) ) {
			self::$instance = new CDNBULL_Backups();
		}

		return self::$instance;

	}

	/**
	 * Setup HM Backup
	 *
	 * @access publics
	 * @see HM_Backup
	 */
	public function __construct() {

		parent::__construct();

		// Set the backup path
		$this->set_path( $this->path() );

		// Set the excludes
		$options = CDNBULL::get_options();
        $options['excluded_strings'] = explode( PHP_EOL, $options['excluded_strings'] );

		$this->set_excludes( apply_filters( 'cdnbull_backup_excludes', CDNBULL_GLOBAL_EXCLUDES ) );
		$this->set_excludes( apply_filters( 'cdnbull_backup_excludes', $options['excluded_strings'] ), true );

		$this->set_includes( apply_filters( 'cdnbull_backup_includes', $options['included_file_extensions'] ) );

		$this->filesize_transient = 'cdnbull_' . '_' . $this->get_type() . '_' . substr( md5( $this->exclude_string() ), 20 ) . '_filesize';

		if ( ! function_exists( 'WP_Filesystem' ) ) {
		    require_once ABSPATH . 'wp-admin/includes/file.php';
		}

		WP_Filesystem();

	}

    public static function backup_request() {

        if ( ! empty( $_REQUEST['_cdn'] ) && sanitize_text_field( wp_unslash( $_REQUEST['_cdn'] ) ) === 'synct' ) {

	    	$backup = CDNBULL_Backups::get_instance()->get_status();

	    	if(empty($backup) ){
                printf(
                    '<div class="notice notice-error is-dismissible"><p class="response">%s</p></div>',
                    esc_html__('Please wait...', 'cdn-bull' )
                );
                add_action( 'admin_footer', array( 'CDNBULL', 'do_backup_javascript' ) );
				add_action( 'admin_footer', array( 'CDNBULL', 'check_sync_javascript' ) );

	    	} else {
	    		wp_safe_redirect( add_query_arg( ['page' => CDNBULL_PLUGIN_SLUG, 'tab' => 'synct',],admin_url('options-general.php') ) );
	    	}

        }
    }

	/**
	 * Perform a backup of the site
	 * @return bool|WP_Error
	 */
	public function do_backup() {

		// Validate nonce
		if ( ! isset( $_REQUEST['_wpnonce'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_REQUEST['_wpnonce'] ) ), 'cdnbull_do_backup' ) ) {
			$status = esc_html__( 'check failed!', 'cdn-bull' );
			$transient_name = 'cdnbull_config_sync_' . get_current_user_id();

	        $response = array(
	             'wrapper' => '<div class="notice notice-error is-dismissible"><p><strong>%1$s</strong> %2$s</p></div>',
	             'subject' => 'Security: ',
	             'message' => $status,
	         );
	         set_transient( $transient_name, $response, 60 * 2);
	         return;
		}

		@ignore_user_abort( true );

		$this->set_status( 'Starting backup...' );

		$this->set_start_timestamp();

		$this->backup();

		if ( ! file_exists( $this->get_archive_filepath() ) ) {

			$errors = $this->get_errors();
			if ( ! empty( $errors ) )
				return new WP_Error( 'backup-failed', implode( ', ', $errors ) );
			else
				return new WP_Error( 'backup-failed', esc_html__( 'Backup file is missing.', 'cdn-bull' ) );

		}

		return true;

	}

	/**
	 * Get the backup once it has run, will return status running as a WP Error
	 *
	 * @return WP_Error|string
	 */
	public function get_backup() {

		global $is_apache;

		// Restore the start timestamp to global scope so HM Backup recognizes the proper archive file
		$this->restore_start_timestamp();

		if ( $status = $this->get_status() ) {

			if ( $this->is_backup_still_running() )
				return new WP_Error( 'error-status', $status );
			else
				return new WP_Error( 'backup-process-killed', esc_html__( 'Backup process failed or was killed.', 'cdn-bull' ) );
		}

		$backup = $backupWithoutKey = $this->get_archive_filepath();

		if ( file_exists( $backup ) ) {

			// Append the secret key on apache servers
			if ( $is_apache && $this->key() ) {

				$backup = add_query_arg( 'key', $this->key(), $backup );

			    // Force the .htaccess to be rebuilt
			    if ( file_exists( $this->get_path() . '/.htaccess' ) )
			        wp_delete_file( $this->get_path() . '/.htaccess' );

			    $this->path();

			}

			$upload_dir = wp_upload_dir();

			$response = new stdClass;
			$response->url = str_replace( parent::conform_dir( $upload_dir['basedir'] ), $upload_dir['baseurl'], $backup );

			$response->path = $backupWithoutKey;
			$response->seconds_elapsed = time() - $this->start_timestamp;
			return $response;

		}

		return new WP_Error( 'backup-failed', esc_html__( 'No backup was found', 'cdn-bull' ) );

	}

	/**
	 * Get the backup once it has run, will return status running as a WP Error
	 *
	 * @return WP_Error|string
	 */
	public static function get_backups() {

		global $is_apache;

		// Restore the start timestamp to global scope so HM Backup recognizes the proper archive file
		$this->restore_start_timestamp();

		if ( $status = $this->get_status() ) {

			if ( $this->is_backup_still_running() )
				return new WP_Error( 'error-status', $status );
			else
				return new WP_Error( 'backup-process-killed', esc_html__( 'Backup process failed or was killed.', 'cdn-bull' ) );
		}

 		$backup = $this->get_path();
 		$upload_dir = wp_upload_dir();

 		if(is_dir($backup)){

 			$bfiles = scandir($backup);

 			if($bfiles){
 				$rfiles = array();
 				$path = get_option( 'cdnbull_backup_path' );

 				foreach($bfiles as $bfile){
 					if(preg_match('/.zip/',$bfile)){
 						$backup = $path. '/'.add_query_arg( 'key', $this->key(), $bfile );

						$rfiles[] = str_replace( parent::conform_dir( wp_upload_dir() ), $upload_dir['baseurl'], $backup );
 					}
 				}

 				rsort($rfiles);

				// Append the secret key on apache servers
				if ( $is_apache && $this->key() ) {

				    // Force the .htaccess to be rebuilt
				    if ( file_exists( $this->get_path() . '/.htaccess' ) )
				        wp_delete_file( $this->get_path() . '/.htaccess' );

				    $this->path();

				}
 			}
 			return $rfiles;
 		}

	return new WP_Error( 'backup-failed', esc_html__( 'No backup was found', 'cdn-bull' ) );
	}

	/**
	 * Remove the backups directoy and everything it contains
	 *
	 * @access public
	 * @return void
	 */
	public function cleanup() {

		$this->rmdir_recursive( $this->get_path() );
		delete_option( 'cdnbull_backup_path' );

	}

	/**
	 * Cleanup old ZipArchive partials that may have been left by old processes
	 */
	public function cleanup_ziparchive($zips) {

		if ( ! file_exists( $this->get_path() . '/'.$zips.'.zip'  ) ) {

				return false;

		}

		    if(wp_delete_file( $this->get_path() . '/'.$zips.'.zip' )){
		    	return  true;
		    } else {
		    	return false;
		    }

	}

	/**
	 * Cleanup old ZipArchive partials that may have been left by old processes
	 */
	public function cleanup_ziparchive_partials() {

		foreach( glob( $this->get_path() . '/*.zip' ) as $ziparchive_partial ) {
			wp_delete_file( $ziparchive_partial );
		}

	}

	/**
	 * Get the estimated size of the sites files and database
	 *
	 * If the size hasn't been calculated yet then it fires an API request
	 * to calculate the size and returns string 'Calculating'
	 *
	 * @access public
	 * @return string $size|Calculating
	 */
	public function get_estimate_size() {

		// Check the cache
		if ( $size = get_transient( $this->filesize_transient ) ) {

			// If we have a number, format it and return
			if ( is_numeric( $size ) )
				return size_format( $size, null, '%01u %s' );

			// Otherwise the filesize must still be calculating
			else
				return esc_html__( 'Calculating', 'cdn-bull' );

		}

		// we dont know the size yet, fire off a remote request to get it for later
		// it can take some time so we have a small timeout then return "Calculating"
		global $cdnbull_noauth_nonce;
		wp_remote_get( add_query_arg( array( 'action' => 'cdnbull_calculate_backup_size', 'backup_excludes' => $this->get_excludes() ), add_query_arg( '_wpnonce', $cdnbull_noauth_nonce, admin_url( 'admin-ajax.php' ) ) ), array( 'timeout' => 0.1, 'sslverify' => false ) );

		return esc_html__( 'Calculating', 'cdn-bull' );

	}

	/**
	 * Hook into the actions fired in HM Backup and set the status
	 *
	 * @param $action
	 */
	protected function do_action( $action ) {

		global $wp_filesystem;

		$this->update_heartbeat_timestamp();

		switch ( $action ) :

			case 'hmbkp_backup_started':

				$this->save_backup_process_id();

			break;

			case 'hmbkp_archive_started' :

				if ( $this->is_using_file_manifest() ){
					/* translators: 1: MySql Dump method, 2: Archive method*/
					$status = sprintf( wp_kses_post( '%1$d files remaining to archive %2$s', 'cdn-bull' ), $this->file_manifest_remaining, '(<code>' . $this->get_archive_method() . '</code>)' );
				} else {
					 /* translators: 1: Archive method*/
					 $status = sprintf( wp_kses_post( 'Creating zip archive %s', 'cdn-bull' ), '(<code>' . $this->get_archive_method() . '</code>)' );
				}

				$this->set_status( $status );

	    	break;

	    	case 'hmbkp_archive_verify_started' :
                /* translators: 1: Archive method*/
	    		$this->set_status( sprintf( wp_kses_post( 'Verifying Zip Archive %s', 'cdn-bull' ), '(<code>' . $this->get_archive_method() . '</code>)' ) );

	    	break;

	    	case 'hmbkp_backup_complete' :

	    		if ( $wp_filesystem->exists( $this->get_schedule_running_path() ) )
	    			wp_delete_file( $this->get_schedule_running_path() );

				$this->clear_backup_process_id();

	    	break;

			case 'hmbkp_error' :

			    if ( $this->get_errors() ) {

			        $file = $this->get_path() . '/.backup_errors';

			        // Delete existing error file if it exists
			        if ( $wp_filesystem->exists( $file ) ) {
			            $wp_filesystem->delete( $file );
			        }

			        // Write new error log
			        $error_data = json_encode( $this->get_errors() );

			        if ( $wp_filesystem->is_writable( dirname( $file ) ) ) {
			            $wp_filesystem->put_contents( $file, $error_data, FS_CHMOD_FILE );
			        }

			    }

		    break;

			case 'hmbkp_warning' :

			    if ( $this->get_warnings() ) {

			        $file = $this->get_path() . '/.backup_warnings';

			        // Delete existing error file if it exists
			        if ( $wp_filesystem->exists( $file ) ) {
			            $wp_filesystem->delete( $file );
			        }

			        // Write new error log
			        $error_data = json_encode( $this->get_warnings() );

			        if ( $wp_filesystem->is_writable( dirname( $file ) ) ) {
			            $wp_filesystem->put_contents( $file, $error_data, FS_CHMOD_FILE );
			        }

			    }

	    	break;

	    endswitch;

	}

	/**
	 * Get the path to the backups directory
	 *
	 * Will try to create it if it doesn't exist
	 * and will fallback to default if a custom dir
	 * isn't writable.
	 *
	 * @access private
	 * @see default_path()
	 * @return string $path
	 */
	private function path() {

	    global $wp_filesystem, $is_apache;

		$path = get_option( 'cdnbull_backup_path' );

	    // If the directory doesn't exist or isn't writable, use the default path
	    if ( ! $path || ( $wp_filesystem->is_dir( $path ) && ! $wp_filesystem->is_writable( $path ) ) ||
	        ( ! $wp_filesystem->is_dir( $path ) && ! $wp_filesystem->is_writable( dirname( $path ) ) ) ) {
	    	$path = $this->path_default();
	    }

		// Create the backups directory if it doesn't exist
	    if ( ! $wp_filesystem->is_dir( $path ) && $wp_filesystem->is_writable( dirname( $path ) ) ) {
	        $wp_filesystem->mkdir( $path, FS_CHMOD_DIR );
	    }

	    // If the path has changed, update the cached value
	    if ( get_option( 'cdnbull_backup_path' ) !== $path ) {
			update_option( 'cdnbull_backup_path', $path );
	    }

	    // Protect against directory browsing by including an index.html file
	    $index = trailingslashit( $path ) . 'index.html';

	    if ( ! $wp_filesystem->exists( $index ) && $wp_filesystem->is_writable( $path ) ) {
	        $wp_filesystem->put_contents( $index, '', FS_CHMOD_FILE );
	    }

	    // Protect the directory with a .htaccess file on Apache servers
	    $htaccess = trailingslashit( $path ) . '.htaccess';

	    if ( $is_apache && function_exists( 'insert_with_markers' ) && ! $wp_filesystem->exists( $htaccess ) && $wp_filesystem->is_writable( $path ) ) {
            /* translators: 1: File name*/
			$contents[]	= '# ' . sprintf( esc_html__( 'This %s file ensures that other people cannot download your backup files.', 'cdn-bull' ), '.htaccess' );
			$contents[] = '';
			$contents[] = '<IfModule mod_rewrite.c>';
			$contents[] = 'RewriteEngine On';
			$contents[] = 'RewriteCond %{QUERY_STRING} !key=' . $this->key();
			$contents[] = 'RewriteRule (.*) - [F]';
			$contents[] = '</IfModule>';
			$contents[] = '';

			insert_with_markers( $htaccess, esc_html__( 'CDN Bull Backup', 'cdn-bull' ), $contents );
		}

	    return parent::conform_dir( $path );
	}


	/**
	 * Return the default backup path
	 *
	 * @access private
	 * @return string $path
	 */
	private function path_default() {
	    global $wp_filesystem;

			$upload_dir = wp_upload_dir();
			$dirname = substr( $this->key(), 0, 10 ) . '-backups';
			$path = parent::conform_dir( trailingslashit( $upload_dir['basedir'] ) . $dirname );

	    // If the backup directory can't be created in upload dir, fallback to uploads
	    if ( ( ( ! $wp_filesystem->is_dir( $path ) && ! $wp_filesystem->is_writable( dirname( $path ) ) ) ||
	           ( $wp_filesystem->is_dir( $path ) && ! $wp_filesystem->is_writable( $path ) ) ) &&
	         strpos( $path, $upload_dir['basedir'] ) === false ) {
				$path = parent::conform_dir( trailingslashit( $upload_dir['basedir'] ) . $dirname );
	    }

			return $path;
	}


	/**
	 * Calculate and generate the private key
	 *
	 * @access private
	 * @return string $key
	 */
	private function key() {

		if ( ! empty( $this->key ) )
			return $this->key;

		$key = array( CDNBULL_PLUGIN_PATH, time() );

		foreach ( array( 'AUTH_KEY', 'SECURE_AUTH_KEY', 'LOGGED_IN_KEY', 'NONCE_KEY', 'AUTH_SALT', 'SECURE_AUTH_SALT', 'LOGGED_IN_SALT', 'NONCE_SALT', 'SECRET_KEY' ) as $constant )
			if ( defined( $constant ) )
				$key[] = constant( $constant );

		shuffle( $key );

		$this->key = md5( serialize( $key ) );
		return $this->key;
	}

	/**
	 * Get the status of the running backup.
	 *
	 * @access public
	 * @return string
	 */
	public function get_status() {
	    global $wp_filesystem;

			$file_path = $this->get_schedule_running_path();

	    // Check if file exists using WP_Filesystem
	    if ( ! $wp_filesystem->exists( $file_path ) ) {
				return '';
	    }

	    // Read file contents using WP_Filesystem
	    $status = $wp_filesystem->get_contents( $file_path );

			return $status;
	}

	/**
	 * Get the path to the backup running file that stores the running backup status
	 *
	 * @access private
	 * @return string
	 */
	private function get_schedule_running_path() {
		return $this->get_path() . '/.backup-running';
	}

	/**
	 * Set the status of the running backup
	 *
	 * @access public
	 * @param string $message
	 * @return void
	 */
	private function set_status( $message ) {
	    global $wp_filesystem;


			$file_path = $this->get_schedule_running_path();

	    // Write to file using WP_Filesystem
	    if ( ! $wp_filesystem->put_contents( $file_path, $message, FS_CHMOD_FILE ) ) {
	        return false;
	    }

	    return true;
	}


	/**
	 * Set the start timestamp for the backup
	 */
	private function set_start_timestamp() {
		global $wp_filesystem;
		$this->start_timestamp = current_time( 'timestamp' );
		$wp_filesystem->put_contents( $this->get_path() . '/.start-timestamp', $this->start_timestamp, FS_CHMOD_FILE );
	}

	/**
	 * Restore the start timestamp for the backup
	 */
	private function restore_start_timestamp() {
		global $wp_filesystem;
		if ( $start_timestamp = $wp_filesystem->get_contents( $this->get_path() . '/.start-timestamp' ) )
			$this->start_timestamp = (int) $start_timestamp;
	}

	/**
	 * Update the heartbeat timestamp to the current time.
	 */
	private function update_heartbeat_timestamp() {
		global $wp_filesystem;
		$wp_filesystem->put_contents( $this->get_path() . '/.heartbeat-timestamp',  time(), FS_CHMOD_FILE );
	}

	/**
	 * Get the heartbeat timestamp.
	 */
	private function get_heartbeat_timestamp() {
	    global $wp_filesystem;

	    $heartbeat = $this->get_path() . '/.heartbeat-timestamp';

	    // Check if file exists and read its content using WP_Filesystem
	    if ( $wp_filesystem->exists( $heartbeat ) ) {
	        return (int) $wp_filesystem->get_contents( $heartbeat );
	    }

	    return false;
	}


	/**
	 * Get the file path to the backup process ID log
	 *
	 * @access private
	 */
	private function get_backup_process_id_path() {
		return $this->get_path() . '/.backup-process-id';
	}

	/**
	 * Get the current backup process ID
	 *
	 * @access private
	 */
	private function get_backup_process_id() {
	    global $wp_filesystem;

	    $file = $this->get_backup_process_id_path();

	    // Check if the file exists and read its contents using WP_Filesystem
	    if ( $wp_filesystem->exists( $file ) ) {
	        return (int) trim( $wp_filesystem->get_contents( $file ) );
	    } else {
	        return false;
	    }
	}


	/**
	 * Save this current backup process ID in case
	 * we need to check later whether it was killed in action
	 *
	 * @access private
	 */
	private function save_backup_process_id() {
	    global $wp_filesystem;
	    $wp_filesystem->put_contents( $this->get_backup_process_id_path(), getmypid() );
	}

	/**
	 * Clear the backup process ID
	 *
	 * @access private
	 */
	private function clear_backup_process_id() {
	    global $wp_filesystem;

	    // Delete the backup process ID file if it exists using WP_Filesystem
	    $file_path = $this->get_backup_process_id_path();
	    if ( $wp_filesystem->exists( $file_path ) ) {
	        $wp_filesystem->delete( $file_path );
	    }
	}


	/**
	 * Whether or not a backup appears to be in progress
	 *
	 * @access private
	 */
	private function is_backup_still_running( $context = 'get_backup' ) {

		// Check whether there's supposed to be a backup in progress
		if ( false === ( $process_id = $this->get_backup_process_id() ) )
			return false;

		// When safe mode is enabled, IREMOTE can't modify max_execution_time
		if ( self::is_safe_mode_active() && ini_get( 'max_execution_time' ) )
			$time_to_wait = ini_get( 'max_execution_time' );
		else
			$time_to_wait = 90;

		// Give heartbeat requests a little bit of time to restart
		if ( 'get_backup' == $context )
			$time_to_wait += 15;

		// If the heartbeat has been modified in the last 90 seconds, we might not be dead
		if ( ( time() - $this->get_heartbeat_timestamp() ) < $time_to_wait )
			return true;

		// Check if there's any file being modified.
		$backup_file_dirs = array( $this->get_path() );

		if ( $this->is_using_file_manifest() ) {
			$backup_file_dirs[] = $this->get_file_manifest_dirpath();
		}

		foreach ( $backup_file_dirs as $backup_file_dir ) {
			$backup_files = glob( $backup_file_dir . '/*' );

			$file_mtimes = array();
			foreach( $backup_files as $backup_file ) {
				$file_mtimes[] = filemtime( $backup_file );
		}
			if ( ! empty( $file_mtimes ) ) {
				$latest_file_mtime = max( $file_mtimes );
				if ( ( time() - $latest_file_mtime ) < $time_to_wait )
				return true;
		}
		}

		return false;
	}

	/**
	 * Check if there's a backup in progress, whether it's running,
	 * and restart it if it's not running
	 *
	 * @todo support checking whether the database should exist
	 */
	public function backup_heartbeat() {

		// Restore the start timestamp to global scope so HM Backup recognizes the proper archive file
		$this->restore_start_timestamp();

		// No process means no backup in progress
		if ( ! $this->get_backup_process_id() )
			return false;

		// No file manifest directory means this wasn't a file manifest approach
		if ( ! is_dir( $this->get_file_manifest_dirpath() ) )
			return false;

		// Check whether there's supposed to be a backup in progress
		if ( $this->get_backup_process_id() && $this->is_backup_still_running( 'backup_heartbeat' ) )
			return false;

		// Uh oh, needs to be restarted
		$this->cleanup_ziparchive_partials();

		$this->save_backup_process_id();

		$this->restart_archive();

	}

	/**
	 * Calculate the size of the backup
	 *
	 * Doesn't account for compression
	 *
	 * @access public
	 * @return string
	 */
	public function get_filesize() {

		$filesize = 0;

		// Only try to calculate once per hour
		set_transient( $this->filesize_transient, 'Calculating', time() + 60 * 60 );

    	// Don't include files if database only
   		if ( $this->get_type() != 'database' ) {

    		// Get rid of any cached filesizes
    		clearstatcache();

			$excludes = $this->exclude_string( 'regex' );

			foreach ( $this->get_files() as $file ) {

				// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
				if ( method_exists( $file, 'isDot' ) && $file->isDot() )
					continue;

				if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
					continue;

			    // Excludes
			    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', parent::conform_dir( $file->getPathname() ) ) ) )
			        continue;

			    $filesize += (float) $file->getSize();

			}

		}

		// Cache for a day
		set_transient( $this->filesize_transient, $filesize, time() + 60 * 60 * 24 );

	}

}

/**
 * Calculate the filesize of the site
 *
 * The calculated size is stored in a transient
 */
function cdnbull_ajax_calculate_backup_size() {

	if ( ! isset( $_GET['_wpnonce'] ) || ! wp_verify_nonce( sanitize_text_field( wp_unslash( $_GET['_wpnonce'] ) ), 'cdnbull_calculate_backup_size' ) ) {
	    exit;
	}

	CDNBULL_Backups::get_instance()->get_filesize();
	exit;
}
add_action( 'wp_ajax_nopriv_cdnbull_calculate_backup_size', 'cdnbull_ajax_calculate_backup_size' );
