<?php

/**
 * Generic file and database backup class
 *
 * @version 2.3
 */
class CDNBULL_HM_Backup {

	/**
	 * The path where the backup file should be saved
	 *
	 * @string
	 * @access private
	 */
	private $path = '';

	/**
	 * The backup type, must be either complete, file or database
	 *
	 * @string
	 * @access private
	 */
	private $type = '';

	/**
	 * The start timestamp of the backup
	 *
	 * @int
	 * @access protected
	 */
	protected $start_timestamp;

	/**
	 * The filename of the backup file
	 *
	 * @string
	 * @access private
	 */
	private $archive_filename = '';

	private $archive_verified = '';

	/**
	 * The path to the zip command
	 *
	 * @string
	 * @access private
	 */
	private $zip_command_path;

	/**
	 * An array of exclude rules
	 *
	 * @array
	 * @access private
	 */
	private $excludes = array();

	/**
	 * An array of include rules
	 *
	 * @array
	 * @access private
	 */
	private $includes = array();

	/**
	 * The path that should be backed up
	 *
	 * @var string
	 * @access private
	 */
	private $root = '';

	/**
	 * Holds the current db connection
	 *
	 * @var resource
	 * @access private
	 */
	private $db;

	/**
	 * An array of all the files in root
	 * excluding excludes and unreadable files
	 *
	 * @var array
	 * @access private
	 */
	private $files = array();

	/**
	 * An array of all the files in root
	 * that match the exclude rules
	 *
	 * @var array
	 * @access private
	 */
	private $excluded_files = array();

	/**
	 * An array of all the files in root
	 * that match the include rules
	 *
	 * @var array
	 * @access private
	 */
	private $included_files = array();

	/**
	 * An array of all the files in root
	 * that are unreadable
	 *
	 * @var array
	 * @access private
	 */
	private $unreadable_files = array();

	/**
	 * Contains an array of errors
	 *
	 * @var mixed
	 * @access private
	 */
	private $errors = array();

	/**
	 * Contains an array of warnings
	 *
	 * @var mixed
	 * @access private
	 */
	private $warnings = array();

	/**
	 * The archive method used
	 *
	 * @var string
	 * @access private
	 */
	private $archive_method = '';

	/**
	 * Whether or not to use a file manifest (more write-intensive)
	 *
	 * @var bool
	 * @access private
	 */
	private $using_file_manifest = false;

	/**
	 * The current file manifest file.
	 *
	 * @access private
	 */
	private $current_file_manifest = false;


	/**
	 * Files of the file manifest that have already been archived
	 *
	 * @access private
	 */
	private $file_manifest_already_archived = array();

	/**
	 * When using the file manifest, the number of files that should be
	 * archived per batch.
	 *
	 * @access private
	 */
	private $file_manifest_per_batch = 200;

	/**
	 * Files remaining to be achived in the file manifest.
	 *
	 * @access protected
	 */
	protected $file_manifest_remaining = 0;

	/**
	 * A ZipArchive instance for this instance
	 */
	private $ziparchive = false;

	/**
	 * A PclZip instance for this instance
	 */
	private $pclzip = false;

	/**
	 * Check whether safe mode is active or not
	 *
	 * @param string $ini_get_callback
	 * @return bool
	 */
	public static function is_safe_mode_active( $ini_get_callback = 'ini_get' ) {

		if ( ( $safe_mode = @call_user_func( $ini_get_callback, 'safe_mode' ) ) && strtolower( $safe_mode ) != 'off' )
			return true;

		return false;

	}

	/**
	 * Check whether shell_exec has been disabled.
	 *
	 * @access public
	 * @static
	 * @return bool
	 */
	public static function is_shell_exec_available() {

		// Are we in Safe Mode
		if ( self::is_safe_mode_active() )
			return false;

		// Is shell_exec or escapeshellcmd or escapeshellarg disabled?
		if ( array_intersect( array( 'shell_exec', 'escapeshellarg', 'escapeshellcmd' ), array_map( 'trim', explode( ',', @ini_get( 'disable_functions' ) ) ) ) )
			return false;

		// Functions can also be disabled via suhosin
		if ( array_intersect( array( 'shell_exec', 'escapeshellarg', 'escapeshellcmd' ), array_map( 'trim', explode( ',', @ini_get( 'suhosin.executor.func.blacklist' ) ) ) ) )
			return false;

		// Can we issue a simple echo command?
		if ( ! @shell_exec( 'echo backupwordpress' ) )
			return false;

		return true;

	}

	/**
	 * Sanitize a directory path
	 *
	 * @param string $dir
	 * @param bool   $recursive (default: false)
	 * @return string $dir
	 */
	public static function conform_dir( $dir, $recursive = false ) {

		// Assume empty dir is root
		if ( ! $dir )
			$dir = '/';

		// Replace single forward slash (looks like double slash because we have to escape it)
		$dir = str_replace( '\\', '/', $dir );
		$dir = str_replace( '//', '/', $dir );

		// Remove the trailing slash
		if ( $dir !== '/' )
			$dir = untrailingslashit( $dir );

		// Carry on until completely normalized
		if ( ! $recursive && self::conform_dir( $dir, true ) != $dir )
			return self::conform_dir( $dir );

		return (string) $dir;

	}

	/**
	 * Sets up the default properties
	 *
	 * @access public
	 */
	public function __construct() {

        // check user role
        if ( !is_admin() || !current_user_can( 'manage_options' )) {
            return;
        }

		add_action('init', [$this, 'load_pclzip']);

		if ( ! function_exists( 'WP_Filesystem' ) ) {
		    require_once ABSPATH . 'wp-admin/includes/file.php';
		}

		WP_Filesystem();

		// Set the error handler (Only if necessary, and avoid in production)
		if ( defined( 'WP_DEBUG' ) && WP_DEBUG ) {
		    set_error_handler( array( &$this, 'error_handler' ) );
		}

	}

	/**
	 * Get the full filepath to the archive file
	 *
	 * @access public
	 * @return string
	 */
	public function get_archive_filepath() {

		return trailingslashit( $this->get_path() ) . $this->get_archive_filename();

	}

	/**
	 * Get the filename of the archive file
	 *
	 * @access public
	 * @return string
	 */
	public function get_archive_filename() {

		if ( empty( $this->archive_filename ) ) {

			if ( empty( $this->start_timestamp ) )
				$this->start_timestamp = current_time( 'timestamp' );

			$this->set_archive_filename( implode( '-', array( sanitize_title( str_ireplace( array( 'http://', 'https://', 'www' ), '', home_url() ) ), 'backup', gmdate( 'Y-m-d-H-i-s', $this->start_timestamp ) ) ) . '.zip' );
		}

		return $this->archive_filename;

	}

	/**
	 * Set the filename of the archive file
	 *
	 * @param string $filename
	 * @throws Exception
	 */
	public function set_archive_filename( $filename ) {

		if ( empty( $filename ) || ! is_string( $filename ) )
			throw new Exception( esc_html( 'archive filename must be a non empty string', 'cdn-bull' ) );

		if ( pathinfo( $filename, PATHINFO_EXTENSION ) !== 'zip' )
			throw new Exception( wp_kses_post( 'invalid file extension for archive filename', 'cdn-bull' ) .  '<code>' . esc_html( $filename ) . '</code>' );

		$this->archive_filename = strtolower( sanitize_file_name( remove_accents( $filename ) ) );

	}

    /**
     * Get the root directory to backup from
     *
     * Defaults to the root of the path equivalent of your home_url
     *
     * @access public
     * @return string
     */
    public function get_root() {

		// scan only current upload dir.
		$upload_dir   = wp_upload_dir();

		if ( empty( $this->root ) )
			$this->set_root( self::conform_dir( $upload_dir['basedir'] ) );

        return $this->root;

    }

	/**
	 * Set the root directory to backup from
	 *
	 * @param string $path
	 * @throws Exception
	 */
	public function set_root( $path ) {

    	if ( empty( $path ) || ! is_string( $path ) || ! is_dir ( $path ) )
    		throw new Exception( sprintf(
	    		/* translators: 1: Directory name*/
	    		wp_kses_post( 'Invalid root path %s must be a valid directory path', 'cdn-bull' ),
	    		'<code>' . esc_html( $path ) . '</code>' )
    		);

    	$this->root = self::conform_dir( $path );

    }

    /**
     * Get the directory backups are saved to
     *
     * @access public
     * @return string
     */
    public function get_path() {

		if ( empty( $this->path ) )
			$this->set_path( self::conform_dir( self::get_path_default() ) );

        return $this->path;

    }

	/**
	 * Get default backup path to save to.
	 *
	 * @return string
	 */
	protected function get_path_default() {
		$upload_dir   = wp_upload_dir();
		return $upload_dir['basedir'] . '/backups/';
	}

	/**
	 * Set the directory backups are saved to
	 *
	 * @param string $path
	 * @throws Exception
	 */
	public function set_path( $path ) {

    	if ( empty( $path ) || ! is_string( $path ) )
    		throw new Exception( sprintf(
    			/* translators: 1: Directory name*/
    			wp_kses_post( 'Invalid backup path %s must be a non empty (string)', 'cdn-bull' ),
    			'<code>' . esc_html( $path ) . '</code>' )
    		);

    	$this->path = self::conform_dir( $path );

    }

	/**
	 * Get the archive method that was used for the backup
	 *
	 * Will be either zip, ZipArchive or PclZip
	 *
	 * @access public
	 */
	public function get_archive_method() {
		return $this->archive_method;
	}

	/**
	 * Whether or not to use the file manifest
	 *
	 * @access public
	 */
	public function is_using_file_manifest() {
		return apply_filters( 'cdnbull_use_file_manifest', (bool)$this->using_file_manifest );
	}

	/**
	 * Set whether or not to use the file manifest
	 *
	 * @access public
	 */
	public function set_is_using_file_manifest( $val ) {
		$this->using_file_manifest = (bool)$val;
	}

	/**
	 * Create a series of file manifests for the backup
	 *
	 * @access private
	 */
	private function create_file_manifests() {

		global $wp_filesystem;

		$dir = $this->get_file_manifest_dirpath();

		if ( $wp_filesystem->is_dir( $dir ) )
			$this->rmdir_recursive( $dir );

		if ( ! $wp_filesystem->is_dir( $dir ) ) {
		    $wp_filesystem->mkdir( $dir, FS_CHMOD_DIR );
		}

		// Protect against directory browsing by including a index.html file
		$index = $dir . '/index.html';
		if ( ! $wp_filesystem->exists( $index ) && $wp_filesystem->is_writable( $dir ) )
			$wp_filesystem->put_contents( $index, '', FS_CHMOD_FILE );

		$excludes = $this->exclude_string( 'regex' );
		$exclude = $this->get_excludes();
		$includes = $this->get_includes();

		$file_manifest = array();
		$this->file_manifest_remaining = 0;
		$file_manifest_file_count = 0;
		$current_batch = 0;
		$current_file = false;
		$line = '';

		foreach( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			// Skip unreadable files
			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
				continue;

		    // Excludes
		    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) )
		        continue;

		    // includes
		    if ( $includes ){
		    	$inc = false;
	            foreach ( $includes as $included_string ) {
	                if ( strpos( str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ), $included_string) !== false ) {
	                    $inc = true;
	                }
	            }

	            if($inc === false)
	            continue;
		    }

		    if ( $exclude ){
		    	$exc = false;
	            foreach ( $exclude as $excluded_string ) {
	                if ( strpos( str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ), $excluded_string ) !== false ) {
	                	$exc = true;
	                }
	            }

	            if($exc === true)
	            continue;
		    }


			if ( $file->isDir() )
				$line .= trailingslashit( str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ). PHP_EOL;

			elseif ( $file->isFile() )
				$line .= str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ). PHP_EOL;

			// File manifest is full
			if ( $current_batch >= $this->file_manifest_per_batch ) {
				$wp_filesystem->put_contents( $current_file, $line . PHP_EOL, FS_CHMOD_FILE );
				$line = '';
				$current_file = false;
			}

			// Create a new file manifest
			if ( $current_file === false ) {
				$file_manifest_file_count++;
				$file_manifest_filename = str_pad( $file_manifest_file_count, 10, "0", STR_PAD_LEFT );
				$current_file = $dir . '/' . $file_manifest_filename . '.txt';
				$current_batch = 0;
			}

			// Write the line to the file manifest if it isn't empty for some reason
			if ( ! empty( $line ) ) {
				$this->file_manifest_remaining++;
				$current_batch++;
			}

		}

		$wp_filesystem->put_contents( $this->get_path() . '/.file-manifest-remaining', $this->file_manifest_remaining, FS_CHMOD_FILE );
		return true;
	}

	/**
	 * Delete the current file manifest
	 *
	 * @access private
	 */
	private function delete_current_file_manifest() {

		if ( ! file_exists( $this->current_file_manifest ) )
			return false;

		// Remove the file manifest because it's already been archived
		wp_delete_file( $this->current_file_manifest );

		// Update the count of remaining files.
		$this->file_manifest_remaining = $this->file_manifest_remaining - count( $this->file_manifest_already_archived );
		if ( $this->file_manifest_remaining < 0 )
			$this->file_manifest_remaining = 0;
		file_put_contents( $this->get_path() . '/.file-manifest-remaining', $this->file_manifest_remaining );

		$this->file_manifest_already_archived = array();

	}


	/**
	 * Get the path to the file manifest directory
	 *
	 * @access private
	 */
	protected function get_file_manifest_dirpath() {
		return $this->get_path() . '/.file-manifests';
	}

	/**
	 * Get batch of files to archive from the file manifest
	 * Ignore any files that already have been archived
	 *
	 * @access private
	 */
	private function get_next_files_from_file_manifest() {

		if ( ! is_dir( $this->get_file_manifest_dirpath() ) )
			return array();

		$files = glob( $this->get_file_manifest_dirpath() . '/*.txt' );
		if ( empty( $files ) )
			return array();

		$this->current_file_manifest = array_shift( $files );

		$files = file_get_contents( $this->current_file_manifest );
		$files = array_map( 'trim', explode( PHP_EOL, $files ) );
		if ( empty( $files ) )
			return array();

		$this->file_manifest_remaining = (int)file_get_contents( $this->get_path() . '/.file-manifest-remaining' );

		return $files;
	}

	/**
	 * Get the backup type
	 *
	 * Defaults to complete
	 *
	 * @access public
	 */
	public function get_type() {

		if ( empty( $this->type ) )
			$this->set_type( 'complete' );

		return $this->type;

	}

	/**
	 * Set the backup type
	 *
	 * $type must be one of complete, database or file
	 * @param string $type
	 * @throws Exception
	 */
	public function set_type( $type ) {

		if ( ! is_string( $type ) || ! in_array( $type, array( 'file', 'complete' ) ) )
			throw new Exception( sprintf(
					/* translators: 1: Backup type parameter*/
					wp_kses_post( 'Invalid backup type %s must be one of (string) file, complete', 'cdn-bull' ),
					'<code>' . esc_html( $type ) . '</code>'
				)
			);

		$this->type = $type;

	}

	/**
	 * Get the path to the zip bin
	 *
	 * If not explicitly set will attempt to work
	 * it out by checking common locations
	 *
	 * @access public
	 * @return string
	 */
	public function get_zip_command_path() {

		// Check shell_exec is available
		if ( ! self::is_shell_exec_available() )
			return '';

		// Return now if it's already been set
		if ( isset( $this->zip_command_path ) )
			return $this->zip_command_path;

		$this->zip_command_path = '';

		// Does zip work
		if ( is_null( shell_exec( 'hash zip 2>&1' ) ) ) {

			// If so store it for later
			$this->set_zip_command_path( 'zip' );

			// And return now
			return $this->zip_command_path;

		}

		// List of possible zip locations
		$zip_locations = array(
			'/usr/bin/zip',
			'/opt/local/bin/zip',
		);

		// Find the one which works
		foreach ( $zip_locations as $location )
			if ( @is_executable( self::conform_dir( $location ) ) )
				$this->set_zip_command_path( $location );

		return $this->zip_command_path;

	}

	/**
	 * Set the path to the zip bin
	 *
	 * Setting the path to false will cause the database
	 * dump to use the php fallback
	 *
	 * @access public
	 * @param mixed $path
	 */
	public function set_zip_command_path( $path ) {

		$this->zip_command_path = $path;

	}

	/**
	 * Set up the ZipArchive instance if ZipArchive is available
	 */
	protected function &setup_ziparchive() {

		// Instance is already open
		if ( ! empty( $this->ziparchive ) ) {
			$this->ziparchive->open( $this->get_archive_filepath(), ZIPARCHIVE::CREATE );
			return $this->ziparchive;
		}

		$ziparchive = new ZipArchive;

		// Try opening ZipArchive
		if ( ! file_exists( $this->get_archive_filepath() ) )
			$ret = $ziparchive->open( $this->get_archive_filepath(), ZIPARCHIVE::CREATE );
		else
			$ret = $ziparchive->open( $this->get_archive_filepath() );

		// File couldn't be opened
		if ( ! $ret )
			return false;

		// Try closing ZipArchive
		$ret = $ziparchive->close();

		// File couldn't be closed
		if ( ! $ret )
			return false;

		// Open it once more
		if ( ! file_exists( $this->get_archive_filepath() ) )
			$ziparchive->open( $this->get_archive_filepath(), ZIPARCHIVE::CREATE );
		else
			$ziparchive->open( $this->get_archive_filepath() );

		$this->ziparchive = $ziparchive;
		return $this->ziparchive;
	}

	/**
	 * Set up the PclZip instance
	 *
	 * @access protected
	 */
	protected function &setup_pclzip() {

		if ( empty( $this->pclzip ) ) {
			$this->pclzip = new PclZip( $this->get_archive_filepath() );
		}
		return $this->pclzip;
	}

	protected function do_action( $action ) {

		do_action( $action, $this );

	}

	/**
	 * Kick off a backup
	 *
	 * @access public
	 * @return bool
	 */
	public function backup() {

		$this->do_action( 'hmbkp_backup_started' );

		// Zip everything up
		$this->archive();

		$this->do_action( 'hmbkp_backup_complete' );

	}

	/**
	 * Zip up all the files.
	 *
	 * Attempts to use the shell zip command, if
	 * thats not available then it falls back to
	 * PHP ZipArchive and finally PclZip.
	 *
	 * @access public
	 */
	public function archive() {

		// If using a manifest, perform the backup in chunks
		if ( 'database' !== $this->get_type()
			&& $this->is_using_file_manifest()
			&& $this->create_file_manifests() ) {

			$this->archive_via_file_manifest();

		} else {

			$this->archive_via_single_request();

		}

	}

	/**
	 * Archive with a file manifest
	 *
	 * @access private
	 */
	private function archive_via_file_manifest() {

		$errors = array();

		// Back up files from the file manifest in chunks
		$next_files = $this->get_next_files_from_file_manifest();
		do {

			$this->do_action( 'hmbkp_archive_started' );

			// `zip` is the most performant archive method
			if ( $this->get_zip_command_path() ) {
				$this->archive_method = 'zip_files';
				$error = $this->zip_files( $next_files );
			}

			// ZipArchive is also pretty fast for chunked backups
			else if ( class_exists( 'ZipArchive' ) && empty( $this->skip_zip_archive ) ) {
				$this->archive_method = 'zip_archive_files';

				$ret = $this->zip_archive_files( $next_files );
				if ( ! $ret ) {
					$this->skip_zip_archive = true;
					continue;
				}
			}

			// Last opportunity
			else {
				$this->archive_method = 'pcl_zip_files';
				$error = $this->pcl_zip_files( $next_files );
			}

			if ( ! empty( $error ) ) {
				$errors[] = $error;
				unset( $error );
			}

			// Update the file manifest with these files that were archived
			$this->file_manifest_already_archived = array_merge( $this->file_manifest_already_archived, $next_files );
			$this->delete_current_file_manifest();

			// Get the next set of files to archive
			$next_files = $this->get_next_files_from_file_manifest();

		} while( ! empty( $next_files ) );

		// If the methods produced any errors, log them
		if ( ! empty( $errors ) )
			$this->warning( $this->get_archive_method(), implode( ', ', $errors ) );

		// ZipArchive has some special reporting requirements
		if ( ! empty( $this->ziparchive ) ) {

			if ( $this->ziparchive->status )
				$this->warning( $this->get_archive_method(), $this->ziparchive->status );

			if ( $this->ziparchive->statusSys )
				$this->warning( $this->get_archive_method(), $this->ziparchive->statusSys );

		}

		// Verify and remove if errors
		$this->verify_archive();

		// Remove the file manifest
		if ( is_dir( $this->get_file_manifest_dirpath() ) )
			$this->rmdir_recursive( $this->get_file_manifest_dirpath() );

		$this->do_action( 'hmbkp_archive_finished' );

	}

	/**
	 * Archive using our traditional method of one request
	 *
	 * @access private
	 */
	private function archive_via_single_request() {

		// Do we have the path to the zip command
		if ( $this->get_zip_command_path() )
			$this->zip();

		// If not or if the shell zip failed then use ZipArchive
		if ( empty( $this->archive_verified ) && class_exists( 'ZipArchive' ) && empty( $this->skip_zip_archive ) )
			$this->zip_archive();

		// If ZipArchive is unavailable or one of the above failed
		if ( empty( $this->archive_verified ) )
			$this->pcl_zip();

		$this->do_action( 'hmbkp_archive_finished' );

	}

	/**
	 * Restart a failed archive process
	 *
	 * @access public
	 */
	public function restart_archive() {

		if ( $this->is_using_file_manifest() ) {

			$this->archive_via_file_manifest();

		} else {

			$this->archive_via_single_request();

		}

		$this->do_action( 'hmbkp_backup_complete' );
	}

	/**
	 * Zip using the native zip command
	 *
	 * @access public
	 */
	public function zip() {

		$this->archive_method = 'zip';

		$this->do_action( 'hmbkp_archive_started' );

		// Zip up $this->root with excludes
		if ( $this->get_type() !== 'database' && $this->exclude_string( 'zip' ) ) {
		    $stderr = shell_exec( 'cd ' . escapeshellarg( $this->get_root() ) . ' && ' . escapeshellcmd( $this->get_zip_command_path() ) . ' -rq ' . escapeshellarg( $this->get_archive_filepath() ) . ' ./' . ' -x ' . $this->exclude_string( 'zip' ) . ' 2>&1' );

		// Zip up $this->root without excludes
		} elseif ( $this->get_type() !== 'database' ) {
		    $stderr = shell_exec( 'cd ' . escapeshellarg( $this->get_root() ) . ' && ' . escapeshellcmd( $this->get_zip_command_path() ) . ' -rq ' . escapeshellarg( $this->get_archive_filepath() ) . ' ./' . ' 2>&1' );

		}

		if ( ! empty( $stderr ) )
			$this->warning( $this->get_archive_method(), $stderr );

		$this->verify_archive();
	}

	/**
	 * Zip one or more files
	 *
	 * @access private
	 */
	private function zip_files( $files ) {

		// Not necessary to include directories when using `zip`
		foreach( $files as $key => $file ) {

			if ( ! is_dir( $file ) )
				continue;

			unset( $files[$key] );
		}

		return shell_exec( 'cd ' . escapeshellarg( $this->get_root() ) . ' && ' . escapeshellcmd( $this->get_zip_command_path() ) . ' ' . escapeshellarg( $this->get_archive_filepath() ) . ' ' . implode( ' ', $files ) . ' -q 2>&1' );
	}

	/**
	 * Fallback for creating zip archives if zip command is
	 * unavailable.
	 */
	public function zip_archive() {

		$this->errors_to_warnings( $this->get_archive_method() );
		$this->archive_method = 'ziparchive';

		$this->do_action( 'hmbkp_archive_started' );

		if ( false === ( $zip = &$this->setup_ziparchive() ) )
			return;

		$excludes = $this->exclude_string( 'regex' );

		if ( $this->get_type() !== 'database' ) {

			$files_added = 0;

			foreach ( $this->get_files() as $file ) {

				// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
				if ( method_exists( $file, 'isDot' ) && $file->isDot() )
					continue;

				// Skip unreadable files
				if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
					continue;

			    // Excludes
			    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) )
			        continue;

			    if ( $file->isDir() )
					$zip->addEmptyDir( trailingslashit( str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) );

			    elseif ( $file->isFile() )
					$zip->addFile( $file->getPathname(), str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) );

				if ( ++$files_added % 500 === 0 )
					if ( ! $zip->close() || ! $zip->open( $this->get_archive_filepath(), ZIPARCHIVE::CREATE ) )
						return;

			}

		}

		if ( $zip->status )
			$this->warning( $this->get_archive_method(), $zip->status );

		if ( $zip->statusSys )
			$this->warning( $this->get_archive_method(), $zip->statusSys );

		$zip->close();

		$this->verify_archive();

	}

	/**
	 * Zip Archive one or more files
	 *
	 * @access private
	 */
	private function zip_archive_files( $files ) {

		if ( false === ( $zip = &$this->setup_ziparchive() ) )
			return false;

		foreach( $files as $file ) {

			$full_path = trailingslashit( $this->get_root() ) . $file;
			if ( is_dir( $full_path ) )
				$zip->addEmptyDir( $file );
			else
				$zip->addFile( $full_path, $file );

		}

		// Avoid limitations in ZipArchive by making sure we save each batch to disk
		$zip->close();
		return true;
	}

	/**
	 * Fallback for creating zip archives if zip command and ZipArchive are
	 * unavailable.
	 *
	 * Uses the PclZip library that ships with WordPress
	 */
	public function pcl_zip() {

		$this->errors_to_warnings( $this->get_archive_method() );
		$this->archive_method = 'pclzip';

		$this->do_action( 'hmbkp_archive_started' );

		global $_cdnbull_hmbkp_exclude_string;

		$_cdnbull_hmbkp_exclude_string = $this->exclude_string( 'regex' );

		$archive = &$this->setup_pclzip();

		// Zip up everything
		if ( $this->get_type() !== 'database' )
			if ( ! $archive->add( $this->get_root(), PCLZIP_OPT_REMOVE_PATH, $this->get_root(), PCLZIP_CB_PRE_ADD, 'cdnbull_hmbkp_pclzip_callback' ) )
				$this->warning( $this->get_archive_method(), $archive->errorInfo( true ) );


		unset( $GLOBALS['_cdnbull_hmbkp_exclude_string'] );

		$this->verify_archive();

	}

	/**
	 * Use PclZip to archive batches of files
	 */
	private function pcl_zip_files( $files ) {

		$this->errors_to_warnings( $this->get_archive_method() );

		$pclzip = &$this->setup_pclzip();

		foreach( $files as $file ) {

			$full_path = trailingslashit( $this->get_root() ) . $file;
			if ( is_dir( $full_path ) )
				continue;

			if ( ! $pclzip->add( $full_path, PCLZIP_OPT_REMOVE_PATH, $this->get_root() ) )
				$this->warning( $this->get_archive_method(), $pclzip->errorInfo( true ) );

		}

	}

	/**
	 * Verify that the archive is valid and contains all the files it should contain.
	 *
	 * @access public
	 * @return bool
	 */
	public function verify_archive() {

		$this->do_action( 'hmbkp_archive_verify_started' );

		// If we've already passed then no need to check again
		if ( ! empty( $this->archive_verified ) )
			return true;

		// If there are errors delete the backup file.
		if ( $this->get_errors( $this->get_archive_method() ) && file_exists( $this->get_archive_filepath() ) )
			wp_delete_file( $this->get_archive_filepath() );

		// If the archive file still exists assume it's good
		if ( file_exists( $this->get_archive_filepath() ) )
			return $this->archive_verified = true;

		return false;

	}

	/**
	 * Return an array of all files in the filesystem
	 *
	 * @access public
	 * @return array
	 */
	public function get_files() {

		if ( ! empty( $this->files ) )
			return $this->files;

		$this->files = array();

		// We only want to use the RecursiveDirectoryIterator if the FOLLOW_SYMLINKS flag is available
		if ( defined( 'RecursiveDirectoryIterator::FOLLOW_SYMLINKS' ) ) {

			$this->files = new RecursiveIteratorIterator( new RecursiveDirectoryIterator( $this->get_root(), RecursiveDirectoryIterator::FOLLOW_SYMLINKS ), RecursiveIteratorIterator::SELF_FIRST, RecursiveIteratorIterator::CATCH_GET_CHILD );

			// Skip dot files if the SKIP_Dots flag is available
			if ( defined( 'RecursiveDirectoryIterator::SKIP_DOTS' ) )
				$this->files->setFlags( RecursiveDirectoryIterator::SKIP_DOTS + RecursiveDirectoryIterator::FOLLOW_SYMLINKS );


		// If RecursiveDirectoryIterator::FOLLOW_SYMLINKS isn't available then fallback to a less memory efficient method
		} else {

			$this->files = $this->get_files_fallback( $this->get_root() );

		}

		return $this->files;

	}

	/**
	 * Fallback function for generating a filesystem
	 * array
	 *
	 * Used if RecursiveDirectoryIterator::FOLLOW_SYMLINKS isn't available
	 *
	 * @access private
	 * @param string $dir
	 * @param array $files. (default: array())
	 * @return array
	 */
	private function get_files_fallback( $dir, $files = array() ) {

	    $handle = opendir( $dir );

	    $excludes = $this->exclude_string( 'regex' );

	    while ( $file = readdir( $handle ) ) :

	    	// Ignore current dir and containing dir
	    	if ( $file === '.' || $file === '..' )
	    		continue;

	    	$filepath = self::conform_dir( trailingslashit( $dir ) . $file );
	    	$file = str_ireplace( trailingslashit( $this->get_root() ), '', $filepath );

	    	$files[] = new SplFileInfo( $filepath );

	    	if ( is_dir( $filepath ) )
	    		$files = $this->get_files_fallback( $filepath, $files );

		endwhile;

		return $files;

	}

	/**
	 * Returns an array of files that will be included in the backup.
	 *
	 * @access public
	 * @return array
	 */
	public function get_included_files() {

		if ( ! empty( $this->included_files ) )
			return $this->included_files;

		$this->included_files = array();

		$excludes = $this->exclude_string( 'regex' );

		foreach ( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			// Skip unreadable files
			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
				continue;

		    // Excludes
		    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) )
		    	continue;

		    $this->included_files[] = $file;

		}

		return $this->included_files;

	}

	/**
	 * Return the number of files included in the backup
	 *
	 * @access public
	 * @return array
	 */
	public function get_included_file_count() {

		if ( ! empty( $this->included_file_count ) )
			return $this->included_file_count;

		$this->included_file_count = 0;

		$excludes = $this->exclude_string( 'regex' );

		foreach ( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			// Skip unreadable files
			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
				continue;

		    // Excludes
		    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) )
		    	continue;

		    $this->included_file_count++;

		}

		return $this->included_file_count;

	}

	/**
	 * Returns an array of files that match the exclude rules.
	 *
	 * @access public
	 * @return array
	 */
	public function get_excluded_files() {

		if ( ! empty( $this->excluded_files ) )
			return $this->excluded_files;

		$this->excluded_files = array();

		$excludes = $this->exclude_string( 'regex' );

		foreach ( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			// Skip unreadable files
			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
				continue;

		    // Excludes
		    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) )
		    	$this->excluded_files[] = $file;

		}

		return $this->excluded_files;

	}

	/**
	 * Return the number of files excluded from the backup
	 *
	 * @access public
	 * @return array
	 */
	public function get_excluded_file_count() {

		if ( ! empty( $this->excluded_file_count ) )
			return $this->excluded_file_count;

		$this->excluded_file_count = 0;

		$excludes = $this->exclude_string( 'regex' );

		foreach ( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			// Skip unreadable files
			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
				continue;

		    // Excludes
		    if ( $excludes && preg_match( '(' . $excludes . ')', str_ireplace( trailingslashit( $this->get_root() ), '', self::conform_dir( $file->getPathname() ) ) ) )
		    	$this->excluded_file_count++;

		}

		return $this->excluded_file_count;

	}

	/**
	 * Returns an array of unreadable files.
	 *
	 * @access public
	 * @return array
	 */
	public function get_unreadable_files() {

		if ( ! empty( $this->unreadable_files ) )
			return $this->unreadable_files;

		$this->unreadable_files = array();

		foreach ( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
			  	$this->unreadable_files[] = $file;

		}

		return $this->unreadable_files;

	}

	/**
	 * Return the number of unreadable files.
	 *
	 * @access public
	 * @return array
	 */
	public function get_unreadable_file_count() {

		if ( ! empty( $this->get_unreadable_file_count ) )
			return $this->get_unreadable_file_count;

		$this->get_unreadable_file_count = 0;

		foreach ( $this->get_files() as $file ) {

			// Skip dot files, they should only exist on versions of PHP between 5.2.11 -> 5.3
			if ( method_exists( $file, 'isDot' ) && $file->isDot() )
				continue;

			if ( ! @realpath( $file->getPathname() ) || ! $file->isReadable() )
			  	$this->get_unreadable_file_count++;

		}

		return $this->get_unreadable_file_count;

	}

	private function load_pclzip() {

		// Load PclZip
		if ( ! defined( 'PCLZIP_TEMPORARY_DIR' ) )
			define( 'PCLZIP_TEMPORARY_DIR', trailingslashit( $this->get_path() ) );

        if (!class_exists('PclZip')) {
            require_once ABSPATH . 'wp-admin/includes/class-pclzip.php';
        }

	}

	/**
	 * Get an array of exclude rules
	 *
	 * The backup path is automatically excluded
	 *
	 * @access public
	 * @return array
	 */
	public function get_excludes() {

		$excludes = array();

		if ( isset( $this->excludes ) )
			$excludes = $this->excludes;

		// If path() is inside root(), exclude it
		if ( strpos( $this->get_path(), $this->get_root() ) !== false )
			array_unshift( $excludes, trailingslashit( $this->get_path() ) );

		return array_unique( $excludes );

	}

	/**
	 * Set the excludes, expects and array
	 *
	 * @access public
	 * @param  Array $excludes
	 * @param Bool $append
	 */
	public function set_excludes( $excludes, $append = false ) {

		if ( is_string( $excludes ) )
			$excludes = explode( ',', $excludes );

		if ( $append )
			$excludes = array_merge( $this->excludes, $excludes );

		$this->excludes = array_filter( array_unique( array_map( 'trim', $excludes ) ) );

	}

	public function get_includes() {

		$includes = array();

		if ( isset( $this->includes ) )
			$includes = $this->includes;

		return array_unique( $includes );

	}

	public function set_includes( $includes, $append = false ) {

		if ( is_string( $includes ) )
			$includes = explode( PHP_EOL, $includes );

		if ( $append )
			$includes = array_merge( $this->includes, $includes );

		$this->includes = array_filter( array_unique( array_map( 'trim', $includes ) ) );

	}

	/**
	 * Generate the exclude param string for the zip backup
	 *
	 * Takes the exclude rules and formats them for use with either
	 * the shell zip command or pclzip
	 *
	 * @access public
	 * @param string $context. (default: 'zip')
	 * @return string
	 */
	public function exclude_string( $context = 'zip' ) {

		// Return a comma separated list by default
		$separator = ', ';
		$wildcard = '';

		// The zip command
		if ( $context === 'zip' ) {
			$wildcard = '*';
			$separator = ' -x ';

		// The PclZip fallback library
		} elseif ( $context === 'regex' ) {
			$wildcard = '([\s\S]*?)';
			$separator = '|';

		}

		$excludes = $this->get_excludes();

		foreach( $excludes as $key => &$rule ) {

			$file = $absolute = $fragment = false;

			// Files don't end with /
			if ( ! in_array( substr( $rule, -1 ), array( '\\', '/' ) ) )
				$file = true;

			// If rule starts with a / then treat as absolute path
			elseif ( in_array( substr( $rule, 0, 1 ), array( '\\', '/' ) ) )
				$absolute = true;

			// Otherwise treat as dir fragment
			else
				$fragment = true;

			// Strip $this->root and conform
			$rule = str_ireplace( $this->get_root(), '', untrailingslashit( self::conform_dir( $rule ) ) );

			// Strip the preceeding slash
			if ( in_array( substr( $rule, 0, 1 ), array( '\\', '/' ) ) )
				$rule = substr( $rule, 1 );

			// Escape string for regex
			if ( $context === 'regex' )
				$rule = str_replace( '.', '\.', $rule );

			// Convert any existing wildcards
			if ( $wildcard !== '*' && strpos( $rule, '*' ) !== false )
				$rule = str_replace( '*', $wildcard, $rule );

			// Wrap directory fragments and files in wildcards for zip
			if ( $context === 'zip' && ( $fragment || $file ) )
				$rule = $wildcard . $rule . $wildcard;

			// Add a wildcard to the end of absolute url for zips
			if ( $context === 'zip' && $absolute )
				$rule .= $wildcard;

			// Add and end carrot to files for pclzip but only if it doesn't end in a wildcard
			if ( $file && $context === 'regex' )
				$rule .= '$';

			// Add a start carrot to absolute urls for pclzip
			if ( $absolute && $context === 'regex' )
				$rule = '^' . $rule;

		}

		// Escape shell args for zip command
		if ( $context === 'zip' )
			$excludes = array_map( 'escapeshellarg', array_unique( $excludes ) );

		return implode( $separator, $excludes );

	}

	/**
	 * Add backquotes to tables and db-names in SQL queries. Taken from phpMyAdmin.
	 *
	 * @param $a_name
	 * @return array|string
	 */
	private function sql_backquote( $a_name ) {

	    if ( ! empty( $a_name ) && $a_name !== '*' ) {

	    	if ( is_array( $a_name ) ) {

	    		$result = array();

	    		reset( $a_name );

	    		while ( list( $key, $val ) = each( $a_name ) )
	    			$result[$key] = '`' . $val . '`';

	    		return $result;

	    	} else {

	    		return '`' . $a_name . '`';

	    	}

	    } else {

	    	return $a_name;

	    }

	}

/**
	 * Get the errors
	 *
	 * @access public
	 */
	public function get_errors( $context = null ) {

		if ( ! empty( $context ) )
			return isset( $this->errors[$context] ) ? $this->errors[$context] : array();

		return $this->errors;

	}

	/**
	 * Add an error to the errors stack
	 *
	 * @access private
	 * @param string $context
	 * @param mixed $error
	 */
	public function error( $context, $error ) {

		if ( empty( $context ) || empty( $error ) )
			return;

		$this->errors[$context][$_key = md5( implode( ':' , (array) $error ) )] = $error;

		$this->do_action( 'hmbkp_error' );

	}

	/**
	 * Migrate errors to warnings
	 *
	 * @access private
	 * @param string $context. (default: null)
	 */
	private function errors_to_warnings( $context = null ) {

		$errors = empty( $context ) ? $this->get_errors() : array( $context => $this->get_errors( $context ) );

		if ( empty( $errors ) )
			return;

		foreach ( $errors as $error_context => $context_errors )
			foreach( $context_errors as $error )
				$this->warning( $error_context, $error );

		if ( $context )
			unset( $this->errors[$context] );

		else
			$this->errors = array();

	}

	/**
	 * Get the warnings
	 *
	 * @access public
	 */
	public function get_warnings( $context = null ) {

		if ( ! empty( $context ) )
			return isset( $this->warnings[$context] ) ? $this->warnings[$context] : array();

		return $this->warnings;

	}

	/**
	 * Add an warning to the warnings stack
	 *
	 * @access private
	 * @param string $context
	 * @param mixed $warning
	 */
	private function warning( $context, $warning ) {

		if ( empty( $context ) || empty( $warning ) )
			return;

		$this->do_action( 'hmbkp_warning' );

		$this->warnings[$context][$_key = md5( implode( ':' , (array) $warning ) )] = $warning;

	}

	/**
	 * Custom error handler for catching php errors
	 *
	 * @param string $type
	 * @return bool
	 */
	public function error_handler( $type ) {

		// Skip strict & deprecated warnings
		if ( ( defined( 'E_DEPRECATED' ) && $type === E_DEPRECATED ) || ( defined( 'E_STRICT' ) && $type === E_STRICT ) || error_reporting() === 0 )
			return false;

		$args = func_get_args();

		array_shift( $args );

		$this->warning( 'php', implode( ', ', array_splice( $args, 0, 3 ) ) );

		return false;

	}

	/**
	 * Recursively delete a directory including
	 * all the files and sub-directories.
	 *
	 * @param string $dir
	 * @return bool
	 */
	public static function rmdir_recursive( $dir ) {
        global $wp_filesystem;

		if ( is_file( $dir ) )
			wp_delete_file( $dir );

	    if ( ! is_dir( $dir ) )
	    	return false;

		$wp_filesystem->delete( $dir, true ); // `true` forces recursive deletion

	}

}

/**
 * Add file callback for PclZip, excludes files
 * and sets the database dump to be stored in the root
 * of the zip
 *
 * @access private
 * @param string $event
 * @param array &$file
 * @return bool
 */
function cdnbull_hmbkp_pclzip_callback( $event, &$file ) {

	global $_cdnbull_hmbkp_exclude_string;

    // Don't try to add unreadable files.
    if ( ! is_readable( $file['filename'] ) || ! file_exists( $file['filename'] ) )
    	return false;

    // Match everything else past the exclude list
    elseif ( $_cdnbull_hmbkp_exclude_string && preg_match( '(' . $_cdnbull_hmbkp_exclude_string . ')', $file['stored_filename'] ) )
    	return false;

    return true;

}
